#ifndef COMMON_DEFINE_H
#define COMMON_DEFINE_H

//#ifndef EIGEN_DONT_ALIGN_STATICALLY
//#define EIGEN_DONT_ALIGN_STATICALLY
//#endif

#include <Eigen/Core> 
#include <Eigen/Dense>
#include <math.h>
#include "iostream"
#include "iomanip"
#include <sys/time.h>

namespace Localization {
namespace Ego {

///======= function ==========//
template <typename Derived>
static Eigen::Matrix<typename Derived::Scalar, 3, 3> SkewSymmetric(const Eigen::MatrixBase<Derived> &q)
{
  Eigen::Matrix<typename Derived::Scalar, 3, 3> ans;
  ans << static_cast<typename Derived::Scalar>(0), -q(2), q(1), q(2), // 2
      static_cast<typename Derived::Scalar>(0), -q(0), -q(1), q(0),
      static_cast<typename Derived::Scalar>(0);
  return ans;
}

template <typename Derived>
static Eigen::Quaternion<typename Derived::Scalar> Positify(const Eigen::QuaternionBase<Derived> &q)
{
  return q;
}

template <typename Derived>
static Eigen::Matrix<typename Derived::Scalar, 4, 4> Qleft(const Eigen::QuaternionBase<Derived> &q) // 4
{
  Eigen::Quaternion<typename Derived::Scalar> qq = Positify(q);
  Eigen::Matrix<typename Derived::Scalar, 4, 4> ans; // 4
  ans(0, 0) = qq.w();
  ans.template block<1, 3>(0, 1) = -qq.vec().transpose();
  ans.template block<3, 1>(1, 0) = qq.vec();
  ans.template block<3, 3>(1, 1) = qq.w() * Eigen::Matrix<typename Derived::Scalar, 3, 3>::Identity() +
                                   SkewSymmetric(qq.vec());
  return ans;
}

static Eigen::Vector3d RotYPR2EulerYPR(const Eigen::Matrix3d &R)
{
  Eigen::Vector3d n = R.col(0);
  Eigen::Vector3d o = R.col(1);
  Eigen::Vector3d a = R.col(2);
  Eigen::Vector3d ypr;
  const double y = atan2(n(1), n(0));
  const double p = atan2(-n(2), n(0) * cos(y) + n(1) * sin(y));
  const double r = atan2(a(0) * sin(y) - a(1) * cos(y), -o(0) * sin(y) + o(1) * cos(y));
  ypr(0) = y;
  ypr(1) = p;
  ypr(2) = r;

  return ypr;
}

static Eigen::Matrix3d EulerYPR2RotYPR(const Eigen::Vector3d &ypr)
{
  double y = ypr(0);
  double p = ypr(1);
  double r = ypr(2);
  Eigen::Matrix3d mRz;
  mRz << cos(y), -sin(y), 0.0, sin(y), cos(y), 0.0, 0.0, 0.0, 1.0;

  Eigen::Matrix3d mRy;
  mRy << cos(p), 0.0, sin(p), 0.0, 1.0, 0.0, -sin(p), 0.0, cos(p);

  Eigen::Matrix3d mRx;
  mRx << 1.0, 0.0, 0.0, 0.0, cos(r), -sin(r), 0.0, sin(r), cos(r);

  return mRz * mRy * mRx;
}

static void ShowSysTime()
{
  struct timeval t;
  gettimeofday(&t,NULL);
  double sec = static_cast<double>(t.tv_sec);
  double usec = static_cast<double>(t.tv_usec) * 1e-6;
  std::cout<<std::setprecision(15)<<"当前系统时间为："<<sec + usec<<std::endl;
}

using Vector3d = Eigen::Vector3d;
using Vector4d = Eigen::Vector4d;
using Matrix3d = Eigen::Matrix3d;
using Matrix4d = Eigen::Matrix4d;

///==============常量===============//
namespace {
    const double GRAVITY             = 9.80665;                 // 重力加速度 G
    const double M_GRAVITY           = 0.00980665;              // mG
    const double U_GRAVITY           = 0.00000980665;           // uG

    const double PI                  = 3.141592653589793;       // pi
    const double PI_2                = 1.570796326794897;       // pi/2
    const double D_PI                = 6.283185307179586;       // 2*pi
    const double RAD2DEG             = 57.295779513082323;      // rad -> deg
    const double DEG2RAD             = 0.017453292519943;       // deg -> rad
    const double KMH2MS              = 1.0 / 3.6;
    const double STATIC_VEL_THESHOLD = 0.002;
    const double HOUR2SEC            = 3600.0;
    const int SAVE_PRECISION         = 15;
    const double bufferLenth         = 0.5;
}

enum StateIndex : uint {
    R = 0,                   
    P = 3,                   
    V = 6,                  
    ACC = 9,                
    AV = 12,                
    BA = 15,                
    BG = 18,               
    IMU_INSTALL_ANGLE = 21,
    WS = 24,              
    STATE_TOTAL = 25
};

enum StateNoiseIndex : uint {
    ACC_NOISE = 0,
    AV_NOISE = 3,
    ACC_RANDOM_WALK = 6,
    GYRO_RANDOM_WALK = 9,
    NOISE_TOTAL = 12
};

enum IndexLLA {
    INDEX_LON = 0,
    INDEX_LAT = 1,
    INDEX_ALT = 2
};

enum IndexVelo {
    INDEX_VE = 0,
    INDEX_VN = 1,
    INDEX_VU = 2
};

enum IndexAtt {
    INDEX_ROLL  = 0,
    INDEX_PITCH = 1,
    INDEX_YAW   = 2
};

enum QuaternionMembers {
    QUAT_W = 0,
    QUAT_X = 1,
    QUAT_Y = 2,
    QUAT_Z = 3
};

enum LocStatus {
    WORK_STATE_PREPARATION      = 0,
    WORK_STATE_ALIGMENT         = 1,
    WORK_STATE_NAV_GOOD         = 2,
    WORK_STATE_NAV_WARNING      = 3,
    WORK_STATE_NAV_DIVERGENCE   = 4,
    WORK_STATE_FAULT            = 5
};

struct State {
    double time;
    Eigen::Quaterniond q;
    Eigen::Vector3d p;
    Eigen::Vector3d v;
    Eigen::Vector3d acc;
    Eigen::Vector3d av;
    Eigen::Vector3d ba;
    Eigen::Vector3d bg;
    double ws;
    Eigen::Quaterniond imuInstallAngle;
    Eigen::MatrixXd mP;
    Eigen::Vector3d att;
    bool isImuValid_;
    bool isVehValid_;
    bool isRtkValid_;
    bool isVlocValid_;
    State() {
      time = 0.0;
      (void)q.setIdentity();
      (void)p.setZero();
      (void)v.setZero();
      (void)acc.setZero();
      (void)av.setZero();
      (void)ba.setZero();
      (void)bg.setZero();
      ws = 1.0;
      (void)imuInstallAngle.setIdentity();
      mP = Eigen::MatrixXd::Zero(StateIndex::STATE_TOTAL, StateIndex::STATE_TOTAL);
      (void)att.setZero();
    }
    void Init() {
      time = 0.0;
      (void)q.setIdentity();
      (void)p.setZero();
      (void)v.setZero();
      (void)acc.setZero();
      (void)av.setZero();
      (void)ba.setZero();
      (void)bg.setZero();
      ws = 1.0;
      (void)imuInstallAngle.setIdentity();
      mP = Eigen::MatrixXd::Zero(StateIndex::STATE_TOTAL, StateIndex::STATE_TOTAL);
      (void)att.setZero();
      isImuValid_ = false;
      isRtkValid_ = false;
      isVehValid_ = false;
      isVlocValid_ = false;
    }
    void Print()
    {
      {
        att = q.matrix().eulerAngles(0,1,2);
        std::cout<<std::setprecision(15)<<"time:"<< time <<std::endl;
        std::cout<<"att(RPY,degree):"<<att.x() * RAD2DEG<<","<<att.y() * RAD2DEG<<","<<att.z() * RAD2DEG<<std::endl;
        std::cout<<"att Matrix:"<<std::endl;
        std::cout<<q.matrix()<<std::endl;
        std::cout<<"q(Y-P-R)(w,x,y,z): "<< q.w()<<","<<q.x()<<","<<q.y()<<","<<q.z()<<std::endl;
        std::cout<<"p-UTM:"<< p.x()<<","<<p.y()<<","<<p.z()<<std::endl;
        std::cout<<"v-UTM:"<<v.x()<<","<<v.y()<<","<<v.z()<<std::endl;
        std::cout<<"acc:"<<acc.x()<<","<<acc.y()<<","<<acc.z()<<std::endl;
        std::cout<<"gyro:"<<av.x()<<","<<av.y()<<","<<av.z()<<std::endl;
        std::cout<<"ba:"<<ba.x()<<","<<ba.y()<<","<<ba.z()<<std::endl;
        std::cout<<"bg:"<<bg.x()<<","<<bg.y()<<","<<bg.z()<<std::endl;
        Eigen::Vector3d yprImuInstallAngel =
            RotYPR2EulerYPR(imuInstallAngle.toRotationMatrix()) / M_PI * 180.0; // 180.0
        std::cout<<"imu install angle YPR from my compute (degree) "<< yprImuInstallAngel.x() <<","<< yprImuInstallAngel.y()
                 <<","<<yprImuInstallAngel.z()<<std::endl;
        std::cout<<"imu install angle Matrix from Eigen (degree):"<<std::endl;
        std::cout<<imuInstallAngle.matrix()<<std::endl;
        std::cout<<"ws:"<<ws<<std::endl;
      }
    }
};

struct LocalizationResult {
    double stamp;
    unsigned char workState         = WORK_STATE_PREPARATION; // align,prepare,good.etc
    unsigned char navMode           = 0U;
    std::uint8_t  integratedState   = 0U;
    unsigned char faultType         = 0U;
    unsigned char localizationState = 0U;
    unsigned short localizationType = 0U;
    double deltRosTime              = 0.0;

    Vector3d posLLA                 = Eigen::Vector3d::Zero();
    Vector3d posUTM                 = Eigen::Vector3d::Zero();
    Vector3d posCov                 = Eigen::Vector3d::Zero();
    Eigen::Quaterniond quat         = Eigen::Quaterniond (1,0,0,0);
    Vector3d attitude               = Eigen::Vector3d::Zero();
    Vector3d attCov                 = Eigen::Vector3d::Zero();
    Vector3d veloENU                = Eigen::Vector3d::Zero();
    Vector3d veloCov                = Eigen::Vector3d::Zero();
    Vector3d angularVelocity        = Eigen::Vector3d::Zero();
    Vector3d avCov                  = Eigen::Vector3d::Zero();
    Vector3d linearAcceleration     = Eigen::Vector3d::Zero();
    Vector3d accCov                 = Eigen::Vector3d::Zero();
    Vector3d accBias                = Eigen::Vector3d::Zero();
    Vector3d accBiasCov             = Eigen::Vector3d::Zero();
    Vector3d gyroBias               = Eigen::Vector3d::Zero();
    Vector3d gyroBiasCov            = Eigen::Vector3d::Zero();
    Vector3d installAng             = Eigen::Vector3d::Zero();
    Vector3d installAngCov          = Eigen::Vector3d::Zero();
    double ws                       = 1.0; // wheel scale
    double wsCov                    = 0.0;

    Eigen::Matrix<double, StateIndex::STATE_TOTAL,StateIndex::STATE_TOTAL> cov;
    bool isImuValid_;
    bool isVehValid_;
    bool isRtkValid_;
    bool isVlocValid_;
    LocalizationResult()
    {
      (void)cov.setZero();
    }

    void Init()
    {
      stamp = 0.0;
      workState       = WORK_STATE_PREPARATION;
      navMode           = 0U;
      integratedState   = 0U;
      faultType         = 0U;
      localizationState = 0U;
      localizationType = 0U;
      deltRosTime              = 0.0;
      attitude               = Eigen::Vector3d::Zero();
      quat                   = Eigen::Quaterniond (1,0,0,0);
      attCov                 = Eigen::Vector3d::Zero();
      posLLA                 = Eigen::Vector3d::Zero();
      posUTM                 = Eigen::Vector3d::Zero();
      posCov                 = Eigen::Vector3d::Zero();
      veloENU                = Eigen::Vector3d::Zero();
      veloCov                = Eigen::Vector3d::Zero();
      linearAcceleration     = Eigen::Vector3d::Zero();
      accCov                 = Eigen::Vector3d::Zero();
      angularVelocity        = Eigen::Vector3d::Zero();
      avCov                  = Eigen::Vector3d::Zero();
      accBias                = Eigen::Vector3d::Zero();
      accBiasCov             = Eigen::Vector3d::Zero();
      gyroBias               = Eigen::Vector3d::Zero();
      gyroBiasCov            = Eigen::Vector3d::Zero();
      installAng             = Eigen::Vector3d::Zero();
      installAngCov          = Eigen::Vector3d::Zero();
      ws                       = 1.0;
      wsCov                    = 0.0;

      isImuValid_ = false;
      isRtkValid_ = false;
      isVehValid_ = false;
      isVlocValid_ = false;
    }
};

} // namespace Ego
} // namespace Localizaiton


#endif
